// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:  none
//
//-----------------------------------------------------------------------------


static const char
rcsid[] = "$Id: s_sound.c,v 1.6 1997/02/03 22:45:12 b1 Exp $";



#include <stdio.h>
#include <stdlib.h>
#include <time.h> //for clock so I know when to restart the midi music with MCI.

#include "i_system.h"
#include "i_sound.h"
#include "sounds.h"
#include "s_sound.h"

#include "z_zone.h"
#include "m_random.h"
#include "w_wad.h"

#include "doomdef.h"
#include "p_local.h"

#include "doomstat.h"
#include "qmus2mid.h"

// Purpose?
const char snd_prefixen[]
= { 'P', 'P', 'A', 'S', 'S', 'S', 'M', 'M', 'M', 'S', 'S', 'S' };

#define S_MAX_VOLUME		127

// when to clip out sounds
// Does not fit the large outdoor areas.
#define S_CLIPPING_DIST		(1200*0x10000)

// Distance to origin when sounds should be maxed out.
// This should relate to movement clipping resolution
// (see BLOCKMAP handling).
// Originally: (200*0x10000).
#define S_CLOSE_DIST		(160*0x10000)


#define S_ATTENUATOR		((S_CLIPPING_DIST-S_CLOSE_DIST)>>FRACBITS)

// Adjustable by menu.
#define NORM_VOLUME    		S_MAX_VOLUME

#define NORM_PITCH     		128
#define NORM_PRIORITY		64
#define NORM_SEP		128

#define S_PITCH_PERTURB		1
#define S_STEREO_SWING		(96*0x10000)

// percent attenuation from front to back
#define S_IFRACVOL		30

#define NA			0
#define S_NUMCHANNELS		2


// Current music/sfx card - index useless
//  w/o a reference LUT in a sound module.
int snd_MusicDevice;
extern int snd_SfxDevice;
// Config file? Same disclaimer as above.
extern int snd_DesiredMusicDevice;
extern int snd_DesiredSfxDevice;



typedef struct
{
    // sound information (if null, channel avail.)
    sfxinfo_t*	sfxinfo;

    // origin of sound
    void*	origin;

    // handle of the sound being played
    int		handle;

} channel_t;


// the set of channels available
static channel_t*	channels;

// These are not used, but should be (menu).
// Maximum volume of a sound effect.
// Internal default is max out of 0-15.
int 		snd_SfxVolume = 127;

// Maximum volume of music. Useless so far.
int 		snd_MusicVolume = 15;



// whether songs are mus_paused
static boolean		mus_paused;

// music currently being played
static musicinfo_t*	mus_playing=0;

// following is set
//  by the defaults code in M_misc:
// number of channels available
int			numChannels;

static int		nextcleanup;



//
// Internals.
//
int
S_getChannel( void*		origin,  sfxinfo_t*	sfxinfo );


int
S_AdjustSoundParams
( mobj_t*	listener,
  mobj_t*	source,
  int*		vol,
  int*		sep,
  int*		pitch );

void S_StopChannel(int cnum);



//
// Initializes sound stuff, including volume
// Sets channels, SFX and music volume,
//  allocates channel buffer, sets S_sfx lookup.
//
void S_Init(int sfxVolume,int musicVolume){
  int		i;

  //Print("S_Init:  Hello.");
  fprintf( stderr, "S_Init: default sfx volume %d\n", sfxVolume);

  // Whatever these did with DMX, these are rather dummies now.
  //I_SetChannels();

  S_SetSfxVolume(sfxVolume);

  S_SetMusicVolume(musicVolume);

  // Allocating the internal channels for mixing
  // (the maximum numer of sounds rendered
  // simultaneously) within zone memory.
  channels =
    (channel_t *) Z_Malloc(numChannels*sizeof(channel_t), PU_STATIC, 0);

  // Free all channels for use
  for (i=0 ; i<numChannels ; i++)
    channels[i].sfxinfo = 0;

  // no sounds are playing, and they are not mus_paused
  mus_paused = 0;

  // Note that sounds have not been cached (yet).
  for (i=1 ; i<NUMSFX ; i++)
    S_sfx[i].lumpnum = S_sfx[i].usefulness = -1;
}




//
// Per level startup code.
// Kills playing sounds at start of level,
//  determines music if any, changes music.
//
void S_Start(void)
{
  int cnum;
  int mnum;

  // kill all playing sounds at start of level
  //  (trust me - a good idea)
  for (cnum=0 ; cnum<numChannels ; cnum++)
    if (channels[cnum].sfxinfo)
      S_StopChannel(cnum);

  // start new music for the level
  mus_paused = 0;

  if (gamemode == commercial)
    mnum = mus_runnin + gamemap - 1;
  else
  {
    int spmus[]=
    {
      // Song - Who? - Where?

      mus_e3m4,	// American	e4m1
      mus_e3m2,	// Romero	e4m2
      mus_e3m3,	// Shawn	e4m3
      mus_e1m5,	// American	e4m4
      mus_e2m7,	// Tim 	e4m5
      mus_e2m4,	// Romero	e4m6
      mus_e2m6,	// J.Anderson	e4m7 CHIRON.WAD
      mus_e2m5,	// Shawn	e4m8
      mus_e1m9	// Tim		e4m9
    };

    if (gameepisode < 4)
      mnum = mus_e1m1 + (gameepisode-1)*9 + gamemap-1;
    else
      mnum = spmus[gamemap-1];
    }

  // HACK FOR COMMERCIAL
  //  if (commercial && mnum > mus_e3m9)
  //      mnum -= mus_e3m9;
  //Print("Play Music!!");
  S_ChangeMusic(mnum, true);

  nextcleanup = 15;
}





/*
ATB
If nobody uses it, why have it?
Shrug.  I'll figure out something here too.
*/
void
S_StartSoundAtVolume(void *origin_p, int sfx_id, int volume){

    int volume2,sep,pitch,chan;
    //sfxinfo_t duh;
    if (sfx_id < 1 || sfx_id > NUMSFX)
        return;

    for (chan=0 ; chan<numChannels ; chan++){
    	if (!channels[chan].sfxinfo)
    	    continue;
    	else if (origin_p &&  channels[chan].origin ==  origin_p)
    	{
    	    S_StopChannel(chan);
    	    break;
    	}
	}

    //if it makes a noise, play it.
    if(S_AdjustSoundParams(players[consoleplayer].mo, origin_p, &volume2, &sep, &pitch)>0)
        chan=addsfx(sfx_id, volume, 0, sep,1);
    else
        return;
    if (S_sfx[sfx_id].usefulness++ < 0)
        S_sfx[sfx_id].usefulness = 1;
    channels[chan].origin=origin_p;
    channels[chan].sfxinfo = &S_sfx[sfx_id];
    channels[chan].handle = woo;

}

void
S_StartSound(void *origin, int sfx_id){
    int volume,sep,pitch,chan;
    //sfxinfo_t duh;

    if (sfx_id < 1 || sfx_id > NUMSFX)
        return;

    for (chan=0 ; chan<numChannels ; chan++){
    	if (!channels[chan].sfxinfo)
    	    continue;
    	else if (origin &&  channels[chan].origin ==  origin)
    	{
    	    S_StopChannel(chan);
    	    break;
    	}
	}

    //if it makes a noise, play it.
    if(S_AdjustSoundParams(players[consoleplayer].mo, origin, &volume, &sep, &pitch)>0)
        chan = addsfx(sfx_id, volume, 0, sep,1);
    else
        return;
    if (S_sfx[sfx_id].usefulness++ < 0)
        S_sfx[sfx_id].usefulness = 1;
    channels[chan].origin=origin;
    channels[chan].sfxinfo= &S_sfx[sfx_id];
    channels[chan].handle = woo;
    //Z_ChangeTag(S_sfx[sfx_id].data, PU_STATIC)
}





void S_StopSound(void *origin)
{

    int cnum;

    for (cnum=0 ; cnum<numChannels ; cnum++)
    {
	if (channels[cnum].sfxinfo && channels[cnum].origin == origin)
	{
	    S_StopChannel(cnum);
	    break;
	}
    }
}









//
// Stop and resume music, during game PAUSE.
//
void S_PauseSound(void)
{
    int i;
    for(i=0;i<NUM_CHANNELS;i++)
    	DS_PauseSound(i);

}

void S_ResumeSound(void)
{
    int i;
    for(i=0;i<NUM_CHANNELS;i++)
    	DS_ResumeSound(i);
}


//
// Updates music & sounds
//
void S_UpdateSounds(void* listener_p)
{
    int		audible;
    int		cnum;
    int		volume;
    int		sep;
    int		pitch;
    int i;
    sfxinfo_t*	sfx;
    channel_t*	c;

    mobj_t*	listener = (mobj_t*)listener_p;



    // Clean up unused data.
    // This is currently not done for 16bit (sounds cached static).
    // DOS 8bit remains.
    if (gametic > nextcleanup)
    {
	for (i=1 ; i<NUMSFX ; i++)
	{
	    if (S_sfx[i].usefulness < 1
		&& S_sfx[i].usefulness > -1)
	    {
		if (--S_sfx[i].usefulness == -1)
		{
		    //if(S_sfx[i].data)
    		//    Z_ChangeTag(S_sfx[i].data, PU_CACHE);
		    //S_sfx[i].data = 0;
		}
	    }
	}
	nextcleanup = gametic + 15;
    }

    for (cnum=0 ; cnum<numChannels ; cnum++){
    	c = &channels[cnum];
    	sfx = c->sfxinfo;
    	if (c->sfxinfo){
            //Print("SoundUpdate");
    	    if (I_SoundIsPlaying(c->handle)){
        		// initialize parameters
        		volume = snd_SfxVolume;
        		pitch = NORM_PITCH;
        		sep = NORM_SEP;

        		if (sfx->link){
        		    pitch = sfx->pitch;
        		    volume += sfx->volume;
        		    if (volume < 1){
            			S_StopChannel(cnum);
            			continue;
        		    }
        		    else if (volume > snd_SfxVolume){
            			volume = snd_SfxVolume;
        		    }
        		}

        		// check non-local sounds for distance clipping
        		//  or modify their params
        		if (c->origin && listener_p != c->origin){
        		    audible = S_AdjustSoundParams(listener,
        						  c->origin,
        						  &volume,
        						  &sep,
        						  &pitch);

        		    if (!audible)
        		    {
        			    S_StopChannel(cnum);
        		    }
        		    else
        			I_UpdateSoundParams(c->handle, volume, sep, pitch);
        		}
    	    }
    	    else{
    		// if channel is allocated but sound has stopped,
    		//  free it
    		    S_StopChannel(cnum);
    	    }

    	}
    }
    // kill music if it is a single-play && finished
    // if (	mus_playing
    //      && !I_QrySongPlaying(mus_playing->handle)
    //      && !mus_paused )
    // S_StopMusic();
}


void S_SetMusicVolume(int volume)
{
    if (volume < 0 || volume > 127)
    {
	I_Error("Attempt to set music volume at %d",
		volume);
    }

    //I_SetMusicVolume(127);
    //I_SetMusicVolume(volume);
    snd_MusicVolume = volume;
}



void S_SetSfxVolume(int volume)
{

    if (volume < 0 || volume > 127)
	I_Error("Attempt to set sfx volume at %d", volume);

    snd_SfxVolume = volume;

}
char errora[256];

int SafemciSendString(char *cmd,char *errora,int len){
    int err;
    err = mciSendString(cmd,errora,255,NULL);
    //gi.bprintf (PRINT_HIGH, "%s\n",errora);
    if (err){
        mciGetErrorString(err,errora,256);
        Print(errora);
    }
    return err;
}
int MusicState=0;
int MusicReplay=0;
int MusicLen =0;
//
// Starts some music with the music id found in sounds.h.
//
void S_StartMusic(int m_id)
{
    S_ChangeMusic(m_id, false);
}

void
S_ChangeMusic
( int			musicnum,
  int			looping )
{
    musicinfo_t*	music;
    char		namebuf[13];
    char        cmd[128];
    int eror,ow=2;
    FILE *damus;

    if(snd_MusicDevice==0){
        //If they have it set up for no music,
        //then don't play any
        return;
    }

    if ( (musicnum <= mus_None)
	 || (musicnum >= NUMMUSIC) )
    {
	//I_Error("Bad music number %d", musicnum);
    	Print("Bad music number");
	    return;
    }
    else
	    music = &S_music[musicnum];

    if (mus_playing == music)
	    return;

    //
    // shutdown old music
	if(MusicState)
	    S_StopMusic();

    //get the mus data.
    if (!music->lumpnum){
        sprintf(namebuf, "d_%s", music->name);
        music->lumpnum = W_GetNumForName(namebuf);
    }
    //get the data.
    music->data = (void *) W_CacheLumpNum(music->lumpnum, PU_MUSIC);
    //open it, write it, close it.
    damus = fopen("atb.mus","wb");
    fwrite(music->data,1,W_LumpLength(music->lumpnum),damus);
    fclose(damus);

    convert( "atb.mus", "atb.mid", 1, 89, 0, 1, &ow ) ;

    sprintf(cmd,"open atb.mid alias music",namebuf);

    eror = SafemciSendString(cmd,errora,256);

    if(eror)
        return;

    sprintf(cmd,"play music from 0");
    eror = SafemciSendString(cmd,errora,256);

    if(eror)
        return;

	MusicState = 1;

    sprintf(cmd,"set music time format ms");
    SafemciSendString(cmd,errora,256);

    sprintf(cmd,"status music length");
    SafemciSendString(cmd,errora,256);

    MusicLen = atol(errora);
    MusicReplay = MusicLen+clock();

    mus_playing = music;
}

void S_CheckMusic(){
    if(MusicState){
        if(clock()>=MusicReplay){
            char cmd[128];
            sprintf(cmd,"play music from 0");
            SafemciSendString(cmd,errora,256);
            MusicReplay = MusicLen+clock();
        }
    }
}


void S_StopMusic(void)
{
    char cmd[128];
    if(MusicState){
        sprintf(cmd,"close music");
        SafemciSendString(cmd,errora,256);
        MusicState = 0;
        MusicReplay=-1;
    }
    mus_playing=0;
}




void S_StopChannel(int cnum)
{

    int		i;
    channel_t*	c = &channels[cnum];

    if (c->sfxinfo)
    {
	// stop the sound playing
    //Print("S_StopChannel 1");

	if (I_SoundIsPlaying(c->handle))
	{
	    //Print("S_StopChannel 2");
	    I_StopSound(c->handle);
	}
	c->sfxinfo->usefulness--;

	c->sfxinfo = 0;
    }
}



//
// Changes volume, stereo-separation, and pitch variables
//  from the norm of a sound effect to be played.
// If the sound is not audible, returns a 0.
// Otherwise, modifies parameters and returns 1.
//
int
S_AdjustSoundParams
( mobj_t*	listener,
  mobj_t*	source,
  int*		vol,
  int*		sep,
  int*		pitch )
{
    fixed_t	approx_dist;
    fixed_t	adx;
    fixed_t	ady;
    angle_t	angle;

    if(source == NULL){
        *vol = NORM_VOLUME;
        *sep = NORM_SEP;
        *pitch = 0;
        return *vol;
    }
    // calculate the distance to sound origin
    //  and clip it if necessary
    adx = abs(listener->x - source->x);
    ady = abs(listener->y - source->y);
    if(adx == 0 && ady == 0){
        *vol = NORM_VOLUME;
        *sep = NORM_SEP;
        *pitch = 0;
        return *vol;
    }

    // From _GG1_ p.428. Appox. eucledian distance fast.
    approx_dist = adx + ady - ((adx < ady ? adx : ady)>>1);

    if (gamemap != 8
	&& approx_dist > S_CLIPPING_DIST)
    {
	return 0;
    }

    // angle of source to listener
    angle = R_PointToAngle2(listener->x,
			    listener->y,
			    source->x,
			    source->y);

    if (angle > listener->angle)
	angle = angle - listener->angle;
    else
	angle = angle + (0xffffffff - listener->angle);

    angle >>= ANGLETOFINESHIFT;

    // stereo separation
    *sep = 128 - (FixedMul(S_STEREO_SWING,finesine[angle])>>FRACBITS);

    // volume calculation
    if (approx_dist < S_CLOSE_DIST)
    {
	*vol = S_MAX_VOLUME;
    }
    else if (gamemap == 8)
    {
	if (approx_dist > S_CLIPPING_DIST)
	    approx_dist = S_CLIPPING_DIST;

	*vol = 127+ ((S_MAX_VOLUME-127)
	    *((S_CLIPPING_DIST - approx_dist)>>FRACBITS))
	    / S_ATTENUATOR;
    }
    else
    {
	// distance effect
	*vol = (S_MAX_VOLUME
		* ((S_CLIPPING_DIST - approx_dist)>>FRACBITS))
	    / S_ATTENUATOR;
    }

    return (*vol > 0);
}




//
// S_getChannel :
//   If none available, return -1.  Otherwise channel #.
//
int
S_getChannel
( void*		origin,
  sfxinfo_t*	sfxinfo )
{
    // channel number to use
    int		cnum;

    channel_t*	c;

    // Find an open channel
    for (cnum=0 ; cnum<numChannels ; cnum++)
    {
    	if (!channels[cnum].sfxinfo)
    	    continue;
    	else if (origin &&  channels[cnum].origin ==  origin)
    	{
    	    S_StopChannel(cnum);
    	    break;
    	}
    }
    if(cnum==numChannels)
        for (cnum=0 ; cnum<numChannels ; cnum++)
        	if (!channels[cnum].sfxinfo)
	            break;

    // None available
    if (cnum == numChannels)
    {
	// Look for lower priority
	for (cnum=0 ; cnum<numChannels ; cnum++)
	    if (channels[cnum].sfxinfo->priority >= sfxinfo->priority) break;

	if (cnum == numChannels)
	{
	    // FUCK!  No lower priority.  Sorry, Charlie.
	    return -1;
	}
	else
	{
	    // Otherwise, kick out lower priority.
	    S_StopChannel(cnum);
	}
    }

    c = &channels[cnum];

    // channel is decided to be cnum.
    c->sfxinfo = sfxinfo;
    c->origin = origin;

    return cnum;
}




